"""A collection of daily updated resources that aren't assigned specific versions."""

from bioversions.utils import DailyGetter


class NCBIGeneGetter(DailyGetter):
    """A getter for NCBI Gene."""

    bioregistry_id = "ncbigene"
    name = "NCBI Gene"



"""A getter for HomoloGene."""

import requests

from bioversions.utils import Getter, VersionType

__all__ = [
    "HomoloGeneGetter",
]

URL = "https://ftp.ncbi.nih.gov/pub/HomoloGene/current/RELEASE_NUMBER"


class HomoloGeneGetter(Getter):
    """A getter for HomoloGene."""

    bioregistry_id = "homologene"
    name = "HomoloGene"
    version_type = VersionType.sequential

    def get(self) -> str:
        """Get the latest HomoloGene version number."""
        s = requests.Session()
        res = s.get(URL, stream=True)
        return res.text.strip()


if __name__ == "__main__":
    HomoloGeneGetter.print()



"""A getter for ROR."""

from bioversions.utils import Getter, VersionType

__all__ = [
    "RORGetter",
]


class RORGetter(Getter):
    """A getter for ROR."""

    bioregistry_id = "ror"
    name = "Research Organization Registry"
    date_fmt = "%Y-%m-%d"
    version_type = VersionType.date

    def get(self) -> dict[str, str]:
        """Get the latest ROR version."""
        import ror_downloader

        version_info = ror_downloader.get_version_info(download=False)
        return {"version": version_info.version, "date": version_info.date.isoformat()}


if __name__ == "__main__":
    RORGetter.print()



"""Get versions from the OLS."""

import logging
from typing import ClassVar, cast

import bioregistry
from bioregistry.external.ols import get_ols_processing
from bioregistry.resolve import get_name
from class_resolver import ClassResolver

from bioversions.utils import Getter, VersionType

logger = logging.getLogger(__name__)

ols_processing = get_ols_processing()


def _get_version_type(bioregistry_id: str) -> VersionType | None:
    ols_id = bioregistry.get_ols_prefix(bioregistry_id)
    if ols_id is None:
        raise ValueError(f"Missing OLS prefix for bioregistry:{bioregistry_id}")
    ols_config = ols_processing.get(ols_id)
    if ols_config is None:
        raise ValueError(
            f"Missing OLS configuration for bioregistry:{bioregistry_id} / ols:{ols_id}"
        )

    ols_version_type = ols_config.version_type
    ols_version_date_format = ols_config.version_date_format
    if ols_version_date_format:
        return VersionType.date
    elif ols_version_type:
        return getattr(VersionType, ols_version_type)
    else:
        logger.warning("[%s] missing version type", bioregistry_id)
        return None


def make_ols_getter(bioregistry_id: str) -> type[Getter] | None:
    """Make a getter from OLS."""
    ols_id = bioregistry.get_ols_prefix(bioregistry_id)
    if ols_id is None:
        return None

    resource = bioregistry.get_resource(bioregistry_id)
    if resource is None:
        logger.warning(f"Invalid bioregistry prefix: {bioregistry_id}")
        return None
    if resource.ols is None:
        logger.warning("[%s] Missing information in OLS", bioregistry_id)
        return None
    version = resource.ols.get("version")
    if version is None:
        logger.debug("[%s] no OLS version", bioregistry_id)
        return None

    _brid = bioregistry_id
    _name = get_name(_brid)
    if _name is None:
        return None
    _version_type = _get_version_type(bioregistry_id)
    if _version_type is None:
        return None

    class OlsGetter(Getter):
        """A getter for OLS data from the Bioregistry."""

        bioregistry_id: ClassVar[str] = _brid
        name: ClassVar[str] = cast(str, _name)
        version_type: ClassVar[str] = _version_type  # type:ignore

        def get(self) -> str:
            """Get the version from the Bioregistry."""
            return cast(str, version)

    return type(f"{_brid.title()}Getter", (OlsGetter,), locals())


def extend_ols(version_getter_resolver: ClassResolver[Getter]) -> None:
    """Add OLS lookup."""
    for bioregistry_id in bioregistry.read_registry():
        try:
            version_getter_resolver.lookup(bioregistry_id)
        except KeyError:
            pass
        else:
            continue
        getter = make_ols_getter(bioregistry_id)
        if getter is None:
            continue
        version_getter_resolver.register(getter)



"""A getter for KEGG."""

from collections.abc import Mapping
from typing import ClassVar

import bioregistry

from bioversions.utils import Getter, VersionType, find, get_soup

__all__ = [
    "KEGGGetter",
]

URL = "https://www.kegg.jp/kegg/docs/relnote.html"


class KEGGGetter(Getter):
    """A getter for KEGG."""

    name = "KEGG"
    date_fmt = "%B %d, %Y"
    version_type = VersionType.semver_minor
    collection: ClassVar[list[str]] = ["kegg", *(bioregistry.get_has_parts("kegg") or [])]

    def get(self) -> Mapping[str, str]:
        """Get the latest KEGG version number."""
        soup = get_soup(URL)
        header = find(soup, "h4")
        sibling = header.next_sibling
        if not sibling:
            raise ValueError
        sibling_text = sibling.text.strip()
        version, date = (part.strip() for part in sibling_text.split(",", 1))
        version = version[len("Release ") :]
        return {"version": version, "date": date}


if __name__ == "__main__":
    KEGGGetter.print()



"""A getter for PubChem."""

import datetime

from bioversions.utils import Getter, VersionType

__all__ = [
    "PubChemGetter",
]


class PubChemGetter(Getter):
    """A getter for PubChem."""

    name = "PubChem"
    homepage_fmt = "https://ftp.ncbi.nlm.nih.gov/pubchem/Compound/Monthly/{version}"
    version_type = VersionType.date

    def get(self):
        """Get the latest PubChem version number."""
        return datetime.datetime.now().strftime("%Y-%m-01")


if __name__ == "__main__":
    PubChemGetter.print()



"""A getter for SILVA."""

from datetime import datetime

import requests

from bioversions.utils import Getter, VersionType

__all__ = [
    "SILVAGetter",
]

URL = "https://www.arb-silva.de/fileadmin/silva_databases/current/VERSION.txt"
HEADER_FMT = "%a, %d %b %Y %H:%M:%S %Z"


class SILVAGetter(Getter):
    """A getter for the SILVA."""

    bioregistry_id = "silva.taxon"
    name = "SILVA ribosomal RNA database"
    version_type = VersionType.sequential
    homepage_fmt = "https://www.arb-silva.de/"

    def get(self):
        """Get the latest SILVA version number from VERSION.txt."""
        res = requests.get(URL, timeout=15)

        # Get version from the content
        rv = {"version": res.text.strip()}

        # Get date from the Last-Modified header
        date_str = res.headers.get("Last-Modified")
        if date_str:
            rv["date"] = datetime.strptime(date_str, HEADER_FMT)

        return rv


if __name__ == "__main__":
    SILVAGetter.print()



"""A getter for CCLE."""

from bioversions.utils import Getter, VersionType

__all__ = [
    "CCLEGetter",
]


class CCLEGetter(Getter):
    """A getter for CCLE."""

    bioregistry_id = "ccle"
    version_type = VersionType.static
    name = "Cancer Cell Line Encylopedia"
    date_fmt = "%Y"

    def get(self):
        """Get the CCLE version number."""
        return {"version": "2019", "date": "2019"}


if __name__ == "__main__":
    CCLEGetter.print()



"""A getter for ChEBI."""

import requests

from bioversions.utils import Getter, VersionType

__all__ = [
    "ChEBIGetter",
]

README = "https://ftp.ebi.ac.uk/pub/databases/chebi/ontology/README"
VERSION_PREFIX = "* ChEBI Release:"
DATE_PREFIX = "* Date of last update:"


class ChEBIGetter(Getter):
    """A getter for ChEBI."""

    bioregistry_id = "chebi"
    name = "ChEBI"
    version_type = VersionType.sequential
    date_fmt = "%Y-%m-%d"

    def get(self):
        """Get the latest ChEBI version number."""
        res = requests.get(README, timeout=5)
        res.raise_for_status()

        rv = {}
        for line in res.iter_lines():
            line = line.decode()
            if line.startswith(VERSION_PREFIX):
                rv["version"] = line.removeprefix(VERSION_PREFIX).strip()
            elif line.startswith(DATE_PREFIX):
                rv["date"] = line.removeprefix(DATE_PREFIX).strip()

        if not rv:
            raise ValueError(f"was not able to parse ChEBI version from {README}")

        return rv


if __name__ == "__main__":
    ChEBIGetter.print()



"""A getter for the OMIM."""

import datetime
from typing import ClassVar

from dateutil.parser import parse

from bioversions.utils import Getter, VersionType, get_soup

__all__ = [
    "OMIMGetter",
]


class OMIMGetter(Getter):
    """A getter for OMIM."""

    name = "Online Mendelian Inheritance in Man"
    version_type = VersionType.date
    collection: ClassVar[list[str]] = ["omim.ps", "omim"]

    def get(self) -> datetime.datetime:
        """Get the latest OMIM version number."""
        soup = get_soup("https://omim.org/")
        for tag in soup.find_all("h5"):
            text = tag.text.strip()
            if text.startswith("Updated"):
                rv = text[len("Updated") :].strip()
                return parse(rv)
        raise ValueError


if __name__ == "__main__":
    OMIMGetter.print()



"""A getter for SPDX."""

from typing import ClassVar

import requests

from bioversions.utils import Getter, VersionType

__all__ = [
    "SPDXGetter",
]

DATA_URL = "https://github.com/spdx/license-list-data/raw/refs/heads/main/json/licenses.json"


class SPDXGetter(Getter):
    """A getter for SPDX."""

    bioregistry_id = "spdx"
    name = "Software Package Data Exchange License"
    version_type = VersionType.other
    collection: ClassVar[list[str]] = ["spdx", "spdx.term"]

    def get(self) -> str:
        """Get the latest SwissLipids version number."""
        return requests.get(DATA_URL, timeout=5).json()["licenseListVersion"]


if __name__ == "__main__":
    SPDXGetter.print()



"""A web application for listing database versions."""

from typing import Any

import flask
from flask_bootstrap import Bootstrap

from bioversions.sources import get_rows, resolve

app = flask.Flask(__name__)
Bootstrap(app)


@app.route("/")
def home() -> str:
    """Show the home page with a list of latest database versions."""
    return flask.render_template("home.html", rows=get_rows())


@app.route("/database/<name>.json")
def database(name: str) -> flask.Response:
    """Resolve information about a given database."""
    rv: dict[str, Any] = {"query": name}
    try:
        bioversion = resolve(name)
    except KeyError:
        rv["success"] = False
    else:
        rv["success"] = True
        rv["result"] = bioversion.model_dump()
    return flask.jsonify(rv)


if __name__ == "__main__":
    app.run()



"""Resources."""

import datetime
import json
from pathlib import Path

import yaml

__all__ = [
    "EXPORT_PATH",
    "VERSIONS_PATH",
    "load_versions",
    "write_export",
    "write_versions",
]

HERE = Path(__file__).parent.resolve()
VERSIONS_PATH = HERE.joinpath(HERE, "versions.json")

ROOT = HERE.parent.parent.parent.resolve()
PYPROJECT_TOML_PATH = ROOT.joinpath("pyproject.toml")
DOCS = ROOT.joinpath("docs")
EXPORTS_DIRECTORY = DOCS.joinpath("_data")
EXPORT_PATH = EXPORTS_DIRECTORY.joinpath("versions.yml")
FAILURES_PATH = DOCS.joinpath("failures.md")


def load_versions():
    """Load Bioversions data."""
    if not VERSIONS_PATH.is_file():
        raise RuntimeError(
            f"bioversions was not packaged/built/installed properly -"
            f"{VERSIONS_PATH.name} was not found inside the distribution"
        )
    with open(VERSIONS_PATH) as file:
        return json.load(file)


def _date_converter(o):
    if isinstance(o, datetime.datetime | datetime.date):
        return o.strftime("%Y-%m-%d")


def write_versions(versions, indent: int = 2, **kwargs) -> None:
    """Write Bioversions data."""
    _ensure_editable()
    with open(VERSIONS_PATH, "w") as file:
        json.dump(versions, file, indent=indent, default=_date_converter, **kwargs)


def write_export(versions) -> None:
    """Write Bioversions data to the export directory."""
    _ensure_editable()
    with open(EXPORT_PATH, "w") as file:
        yaml.safe_dump(versions, file)


def _ensure_editable() -> None:
    if not PYPROJECT_TOML_PATH.is_file():
        raise RuntimeError("can not make export when bioversions is not installed in editable mode")



"""A getter for PathBank."""

from bioversions.utils import Getter, VersionType, find, get_soup

__all__ = [
    "PathBankGetter",
]

URL = "https://pathbank.org/"


class PathBankGetter(Getter):
    """A getter for PathBank."""

    bioregistry_id = "pathbank"
    name = "PathBank"
    version_type = VersionType.semver_minor

    def get(self) -> str:
        """Get the latest PathBank version number."""
        soup = get_soup(URL)
        main = find(soup, id="main")
        footer = find(main, name="footer")
        clear = find(footer, **{"class": "wishart-clear"})
        strong = find(clear, name="strong")
        return strong.text


if __name__ == "__main__":
    PathBankGetter.print()



"""A getter for MeSH."""

import datetime
import ftplib
import logging

from bioversions.utils import Getter, VersionType

__all__ = [
    "MeshGetter",
]

logger = logging.getLogger(__name__)

#: In 2024 and before, these were .gz, but after
#: became .xml files
SUFFIXES = [".xml", ".xml.gz", ".gz"]


class MeshGetter(Getter):
    """A getter for MeSH."""

    bioregistry_id = "mesh"
    name = "MeSH"
    homepage_fmt = "ftp://nlmpubs.nlm.nih.gov/online/mesh/{version}"
    version_type = VersionType.year

    def get(self):
        """Get the latest MeSH version number."""
        try:
            with ftplib.FTP("nlmpubs.nlm.nih.gov") as ftp:
                ftp.login()
                ftp.cwd("/online/mesh/MESH_FILES/xmlmesh/")
                names = [name for name, _ in ftp.mlsd()]

        except Exception as e:
            logger.warning(
                "could not look up MeSH version, falling back to current year. exception: %s", e
            )
            return str(datetime.date.today().year)

        for name in names:
            for suffix in SUFFIXES:
                if name.startswith("desc") and name.endswith(suffix):
                    return name[len("desc") : -len(suffix)]

        raise ValueError("unable to parse results returned from MeSH FTP")


if __name__ == "__main__":
    MeshGetter.print()



"""A getter for WikiPathways."""

from bioversions.utils import Getter, VersionType, get_soup

__all__ = [
    "WikiPathwaysGetter",
]

URL = "http://data.wikipathways.org/current/gmt/"


class WikiPathwaysGetter(Getter):
    """A getter for WikiPathways."""

    bioregistry_id = "wikipathways"
    name = "WikiPathways"
    homepage_fmt = "http://data.wikipathways.org/{version}/"
    date_version_fmt = "%Y%m%d"
    version_type = VersionType.date

    def get(self):
        """Get the latest WikiPathways version number."""
        soup = get_soup(URL)
        if soup is None:
            raise ValueError(f"could not get WikiPathways data from {URL}")
        anchor = soup.find(id="File")
        return anchor.text.split("-")[1]


if __name__ == "__main__":
    WikiPathwaysGetter.print()
